#include "../SDK/foobar2000.h"

#include "wavpack/wavpack.h"

class input_wavpack : public input
{
private:
	WavpackContext *wpc;
	long *sample_buffer;
	void format_samples (int bps, uchar *dst, long *src, ulong samcnt);
public:

	input_wavpack()
	{
		sample_buffer = NULL;
		wpc = NULL;
	}

	~input_wavpack()
	{
		if (wpc) {
			if (WavpackGetNumErrors (wpc))
				console::error (uStringPrintf ("crc error(s) detected in %d blocks",
					WavpackGetNumErrors (wpc)));

		    WavpackCloseFile (wpc);
		}

		if (sample_buffer)
			free (sample_buffer);
	}

	virtual bool test_filename(const char * fn,const char * ext) 
	{
		return !stricmp_utf8(ext,"WV");
	}

	virtual bool open (reader *r, file_info *info, unsigned flags)
	{
		char error [256];

		wpc = WavpackOpenFileInput (r, info->get_file_path (), error, OPEN_WVC | OPEN_NORMALIZE);

		if (!wpc) {
			console::error (error);
			return 0;
		}

		if (flags & OPEN_FLAG_GET_INFO)
		{
			tag_reader::g_run_multi(r,info,"ape|id3v1");

			double dur = (double) WavpackGetNumSamples (wpc) / (double) WavpackGetSampleRate (wpc);
			info->set_length (dur);
			info->info_set_int ("samplerate", WavpackGetSampleRate (wpc));
			info->info_set_int ("channels", WavpackGetNumChannels (wpc));
			info->info_set_int ("bitspersample", WavpackGetBitsPerSample (wpc));
			info->info_set ("format", (WavpackGetMode (wpc) & MODE_FLOAT) ? "ieee float" : "integer");
			info->info_set ("codec", "wavpack");
			info->info_set_int ("bitrate", (int) (WavpackGetAverageBitrate (wpc) / 1000.0));

	    	if (WavpackGetMode (wpc) & MODE_LOSSLESS)
				info->info_set ("compression", (WavpackGetMode (wpc) & MODE_HYBRID) ? "hybrid lossless" : "lossless");
			else
				info->info_set ("compression", (WavpackGetMode (wpc) & MODE_HYBRID) ? "hybrid lossy" : "lossy");
		}

		return 1;
	}

	virtual int run (audio_chunk *chunk)
	{
		int srate, nch, samples;

		if (!sample_buffer)
			sample_buffer = (long *) malloc (576 * sizeof (audio_sample) * WavpackGetNumChannels (wpc));

		samples = WavpackUnpackSamples (wpc, sample_buffer, 576);

		if (!samples)
			return 0;

		nch = WavpackGetNumChannels (wpc);
		srate = WavpackGetSampleRate (wpc);

		if (WavpackGetMode (wpc) & MODE_FLOAT) {
			if (audio_sample_size == 64) {
			    int cnt = samples * nch;
			    double *dst = ((double *) sample_buffer) + cnt;
				float *src = ((float *) sample_buffer) + cnt;

			    while (cnt--)
				    *--dst = *--src;
			}

			chunk->set_data ((audio_sample *) sample_buffer, samples, nch, srate);
		}
		else {
			int bps = WavpackGetBytesPerSample (wpc);

			format_samples (bps, (uchar *) sample_buffer, sample_buffer, samples * nch);
			chunk->set_data_fixedpoint (sample_buffer, samples * bps * nch, srate, nch, bps * 8);
	    }

		return 1;
	}

	virtual bool seek (double seconds)
	{
		if (WavpackSeekSample (wpc, (ulong)(seconds * WavpackGetSampleRate (wpc))))
			return 1;
		else
			return 0;
	}

	virtual bool can_seek () { return 1; }

	virtual input::set_info_t set_info(reader *r,const file_info * info)
	{
		tag_remover::g_run(r);
		return tag_writer::g_run(r,info,"ape") ? SET_INFO_SUCCESS : SET_INFO_FAILURE;
	}

};

void input_wavpack::format_samples (int bps, uchar *dst, long *src, ulong samcnt)
{
	long temp;

	switch (bps) {

		case 1:
			while (samcnt--)
				*dst++ = (uchar)(*src++ + 128);

			break;

		case 2:
			while (samcnt--) {
				*dst++ = (uchar)(temp = *src++);
				*dst++ = (uchar)(temp >> 8);
			}

			break;

		case 3:
			while (samcnt--) {
				*dst++ = (uchar)(temp = *src++);
				*dst++ = (uchar)(temp >> 8);
				*dst++ = (uchar)(temp >> 16);
			}

			break;
    }
}

static service_factory_t<input,input_wavpack> foo;
DECLARE_COMPONENT_VERSION("WavPack Decoder", "2.0b2", NULL);
DECLARE_FILE_TYPE("WavPack files","*.WV");
